#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Command Line Interface."""
from argparse import ArgumentParser
from sys import stderr
from copy import copy

from snap_tam.aggregate_filesystem import run


def _required_fields(parser):
    """Add required fields to parser.

    Args:
        parser (ArgumentParser): Argparse object.

    Returns:
        ArgumentParser: Argeparse object.
    """
    return parser


def _optional_fields(parser):
    """Add optional fields.

    Args:
        parser (ArgumentParser): Argparse object.

    Returns:
        ArgumentParser: Argeparse object.
    """
    parser.add_argument(
        '-p', '--onedrive-path', help='Path to OneDrive installation.')

    return parser


def _add_arguments(parser):
    """Add arguments to parser.

    Args:
        parser (ArgumentParser): Argparse object.

    Returns:
        ArgumentParser: Argeparse object.
    """
    parser2 = _optional_fields(parser)
    parser3 = _required_fields(parser2)

    return parser3


def cli():
    """Command line interface for package.

    Side Effects: Executes program.
    """
    prog_desc = 'Aggregate SNaP reports on OneDrive.'
    argeparser = ArgumentParser(description=prog_desc)
    parser = _add_arguments(copy(argeparser))
    args = parser.parse_args()

    try:
        run(onedrive_dir_path=args.onedrive_path)
    except Exception as err:
        print(err, file=stderr)


if __name__ == '__main__':
    cli()



"""SNaP OneDrive Report Aggregator

# Resources
1. Docs: https://docs.microsoft.com/en-us/onedrive/developer/rest-api/
getting-started/msa-oauth?view=odsp-graph-online

# Minor to-do's
- Verified exists: If need get code and user pastes, does script correctly run
  all the way through without any intermediate error and need to re-run a
  second time? When I was developing, I had to run twice in a row one time.
- Alternative to have user paste this into browser and copy/paste the token?
- how make code and token last longer?:
  - https://docs.microsoft.com/en-us/onedrive/developer/rest-api/getting-started/graph-oauth?view=odsp-graph-online#step-3-get-a-new-access-token-or-refresh-token
  - can update token max life via powershell only
"""
import base64
# import json
import os
import re

import requests
import urllib.parse as urlparse
from urllib.parse import parse_qs
from typing import Dict, List


# Variables
ROOT_DIR = os.path.join(
    os.path.dirname(os.path.realpath(__file__)), '..')
CODE_FILE_PATH = os.path.join(ROOT_DIR, 'code.env')
TOKEN_FILE_PATH = os.path.join(ROOT_DIR, 'token.env')
CONFIG = {
    'redirect_uri': 'https://login.live.com/oauth20_desktop.srf',
    'client_id': os.getenv('ATTEMPT2_APP_ID'),
    'client_secret': os.getenv('n/a', 'wCv35Dzw5AFgOMbE._CJ-EO-4C1-jfFVJH'),
    'api_base_url': 'https://api.onedrive.com/v1.0/',
    'scopes': ['Files.ReadWrite.All'],
    'fetch_headers': {
        'Host': 'graph.microsoft.com',
        'Authorization': 'Bearer {}'
    }
}

# Update config
client_id = CONFIG['client_id']
scope = CONFIG['scopes'][0]
redirect_uri = CONFIG['redirect_uri']
auth_url_base = \
    'https://login.microsoftonline.com/common/oauth2/v2.0/authorize'
fetch_url_base = 'https://graph.microsoft.com/v1.0/'
auth_url = f'{auth_url_base}?client_id={client_id}' \
           f'&scope={scope}&response_type=code&redirect_uri={redirect_uri}'
code_token_headers = {
    'client_id': CONFIG['client_id'],
    'redirect_uri': CONFIG['redirect_uri'],
    'code': None,
    'grant_type': 'authorization_code'
}

CONFIG['auth_url'] = auth_url
CONFIG['code_token_url'] = auth_url
CONFIG['code_token_headers'] = code_token_headers


def get_code(auth_url: str) -> str:
    """Get code"""
    # 1. Auth: Get code
    # - https://docs.microsoft.com/en-us/onedrive/developer/rest-api/getting-started/graph-oauth?view=odsp-graph-online#step-1-get-an-authorization-code
    msg = '2. In "step 1", you should either be asked to ' \
          'log in or auto-logged-in. After that, you should be redirected to ' \
          'a new blank page, and the URL in the address bar should look ' \
          'similar to this: \nhttps://login.live.com/oauth20_desktop.srf?' \
          'code=M.R3_BL2.64370af9-d97a-0e5c-89f8-458184fe1ba8&lc=1033\n\n' \
          'Copy and paste that entire URL here: '
    print('1. Open URL in browser to get auth code: \n', auth_url, '\n')
    url = input(msg)
    parsed = urlparse.urlparse(url)
    code = parse_qs(parsed.query)['code'][0]
    if not code:
        raise RuntimeError('No code found. Url was: ', url)
    return code


def get_token(headers: Dict) -> str:
    """Get token"""
    # 2. redeem token from code
    # - https://docs.microsoft.com/en-us/onedrive/developer/rest-api/getting-started/graph-oauth?view=odsp-graph-online#step-2-redeem-the-code-for-access-tokens
    response: Dict = requests.post(
        'https://login.microsoftonline.com/common/oauth2/v2.0/token',
        data=headers).json()
    return response['access_token']


def download_files(headers: Dict):
    """Download files"""
    # 3. fetch resources
    # - https://docs.microsoft.com/en-us/onedrive/developer/rest-api/api/drive_sharedwithme?view=odsp-graph-online
    # 1. List root files
    # endpoint = 'me/drive/sharedWithMe'
    endpoint = 'me/drive/root/children'
    fetch_url = fetch_url_base + endpoint
    root_items: Dict = requests.get(fetch_url, headers=headers).json()
    if 'error' in root_items.keys():
        raise RuntimeError(root_items)
    root_objs: List[Dict] = root_items['value']
    root_objs: Dict = {
        val['name']: val
        for val in root_objs
    }
    snap_folders = {
        k: v
        for k, v in root_objs.items()
        if re.match('SNAP_TAM_[0-9]{3}', k)
    }
    # Try 1: list & dl (fail)
    # 2. List SNAP folder contents
    # https://docs.microsoft.com/en-us/onedrive/developer/rest-api/api/driveitem_list_children?view=odsp-graph-online
    # snap_folder_contents: Dict = {}
    # endpoint = 'me/drive/items'
    # fetch_url_prefix = fetch_url_base + endpoint
    # for k, v in snap_folders.items():
    #     fetch_url = fetch_url_prefix + '/{}/children'.format(v['id'])
    #     children = requests.get(fetch_url, headers=headers).json()
    #     snap_folder_contents[k] = children
    # 3. Dl files

    # Try 2: straight dl (fail)
    # endpoint = 'me/drive/items/'
    # fetch_url_prefix = fetch_url_base + endpoint
    # for k, v in snap_folders.items():
    #     fetch_url = fetch_url_prefix + '/{}/content'.format(v['id'])
    #     response = requests.get(fetch_url, headers=headers).json()

    # print('Fetch: ' + fetch_url + '\n')
    # pprint(json.dumps(response))

    # Try 3.5: base 64
    # {'error': {'code': 'invalidRequest', 'message': 'Invalid shares key.'
    # for k, v in snap_folders.items():
    #     weburl = v['remoteItem']['webUrl']
    #     ascii_url = weburl.encode('ascii')
    #     base64_url = base64.b64encode(ascii_url)
    #     base64_url = base64_url[0:-1]
    #     # cant replace bytes w/ str:
    #     # encoded_url = base64_url.replace('/', '_').replace('+', '-')
    #     encoded_url = str(base64_url)[2:-1]
    #     fetch_url = fetch_url_prefix + '/{}' \
    #         .format(encoded_url)

    # Try 3: Shares endpoint
    # GET /shares/{shareIdOrUrl}/driveItem?$expand=children
    # - id: {'error': {'code': 'invalidRequest', 'message': 'Bad Argument',
    # 'innerError': ...
    # - remoteItem.id: {'error': {'code': 'invalidRequest', 'message':
    # 'Bad Argument', 'innerError': ...
    # - remoteItem.webUrl: {'error': {'code': 'BadRequest', 'message':
    # "Resource not found for the segment '1drv.ms'.", 'innerError': ...
    # - webUrl: {'error': {'code': 'BadRequest', 'message':
    # "Resource not found for the segment '1drv.ms'.", 'innerError': ...
    # ...
    # 1drv.ms is from webUrls, e.g. https://1drv.ms/u/s!ANBLWsRA6bDCvEk
    snap_folder_contents: Dict = {}
    endpoint = 'shares'
    fetch_url_prefix = fetch_url_base + endpoint
    for k, v in snap_folders.items():
        # fetch_url = fetch_url_prefix + '/{}/driveItem?$expand=children' \
        fetch_url = fetch_url_prefix + '/{}' \
            .format(v['remoteItem']['id'])
            # .format(v['remoteItem']['webUrl'])
        children = requests.get(fetch_url, headers=headers).json()
        snap_folder_contents[k] = children

    print()


def run(config=CONFIG):
    """Run"""
    try:
        headers = config['fetch_headers']
        with open(TOKEN_FILE_PATH, 'r') as file:
            token = file.read()
        headers['Authorization'] = headers['Authorization'].format(token)
        download_files(headers)
    except (RuntimeError, FileNotFoundError):
        # 1. code
        code = get_code(config['auth_url'])
        with open(CODE_FILE_PATH, 'w') as file:
            file.write(code)
        headers = config['code_token_headers']
        headers['code'] = code
        # 2. token
        token = get_token(headers)
        with open(TOKEN_FILE_PATH, 'w') as file:
            file.write(token)
        headers = config['fetch_headers']
        # headers['Authorization'] = headers['Authorization'].format(token)
        headers['Authorization'] = 'Bearer ' + token
        # 3. download
        download_files(headers)


if __name__ == '__main__':
    run()



"""Setup for Pypi"""
import os
from setuptools import setup, find_packages


version = '1.0.3'
packages = find_packages(exclude=['test'])


def get_pkg_data(pkg_name, data_dirs, extensions):
    """Recusrively get package data.

    This should be called from within setup.py's setuptools.setup() function,
    e.g. package_data={'my_package': get_pkg_data(pkg_name='my_pckage', ...)}

    Args:
        pkg_name (str): Name of package containing package data.
        data_dirs (list):
        extensions (list):

    Returns:
        list: List of glob strings where each string represents a terminal
        branch node (i.e. folder) and a glob of all files ending in designated
        extension(s).

    Examples:
        Args: get_pkg_data(pkg_name='my_package',
        src_dirs=['templates'], extensions=['html'])

        Returns: [
        'templates/*.html',
        'templates/content/*.html',
        'templates/content/prompt/*.html',
        'templates/content/prompt/inputs/*.html']
    """
    pkg_root = \
        os.path.dirname(os.path.realpath(__file__)) + '/{}/'.format(pkg_name)
    pkg_data = []
    for _dir in data_dirs:
        for ext in extensions:
            for i, j, y in os.walk(pkg_root + _dir):
                pkg_data.append(i + '/*.' + ext)
    return pkg_data


setup(
    name='snap-tam',
    version=version,
    author='Joe Flack',
    author_email='jflack@jhu.edu',
    url='http://www.modeltb.org',
    packages=packages,
    package_dir={
        'snap_tam': 'snap_tam'
    },
    package_data={
        'snap-tam': [
            '../config.json'
        ]
    },
    entry_points={
        'console_scripts': [
            'snap-tam=snap_tam.cli:cli'
        ],
    },
    # license='LICENSE.txt',
    license='',
    description='Utility for ModelTB.org team. Aggregates SNaP reports to a '
                'single file.',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    install_requires=[
        'openpyxl>=3.0.5',
        'pandas>=1.1.5'
    ]
)



"""SNaP OneDrive Report Aggregator

# Running
1. Replace onedrive_dir_path with your path
2. Run Python script

# Troubleshooting problems
1. "Permission denied"
Solution: Make sure the OneDrive program is running on your computer.

"""
import json
import os
import pandas as pd
import re

from datetime import datetime
from openpyxl import load_workbook
from typing import List, Dict


# Constants
SCRIPT_PATH = os.path.realpath(__file__)
ROOT_DIR = os.path.join(os.path.dirname(SCRIPT_PATH), '..')
CONFIG_PATH = os.path.join(ROOT_DIR, 'config.json')
VALID_EXTENSIONS = ['.xls', '.xlsx', '.xlsm']


def run(
    onedrive_dir_path: str = None,
    input_dir_regexp: str = None,
    worksheet_name: str = None,
    duration_colnames: str = None,
    output_dirname: str = None,
    print_progress: bool = None
):
    """Run"""
    args: Dict = locals()
    config: Dict = None

    with open(CONFIG_PATH, 'r') as config_file:
        config = json.load(config_file)
        for arg_key, arg_val in args.items():
            if arg_key in config.keys() and arg_val:
                config[arg_key] = arg_val
    with open(CONFIG_PATH, 'w') as config_file:
        json_string = json.dumps(
            config,
            default=lambda o: o.__dict__,  # pretty prints
            indent=2)
        config_file.write(json_string)

    onedrive_dir_path: str = config['onedrive_dir_path']
    input_dir_regexp: str = config['input_dir_regexp']
    worksheet_name: str = config['worksheet_name']
    duration_colnames: str = config['duration_colnames']
    output_dirname: str = config['output_dirname']
    print_progress: bool = config['print_progress']

    if not onedrive_dir_path:
        onedrive_dir_path: str = input('Enter path to OneDrive folder: ')
        config['onedrive_dir_path']: str = onedrive_dir_path
        with open(CONFIG_PATH, 'w') as config_file:
            json_string = json.dumps(
                config,
                default=lambda o: o.__dict__,  # pretty prints
                indent=2)
            config_file.write(json_string)


    output_path = os.path.join(onedrive_dir_path, output_dirname)
    onedrive_root_contents: List[str] = os.listdir(onedrive_dir_path)
    snap_dirs: List[str] = [
        os.path.join(onedrive_dir_path, x) for x in onedrive_root_contents
        if re.match(input_dir_regexp, x)]

    snap_files: List[str] = []
    for folder in snap_dirs:
        candidates = [
            y for y in os.listdir(folder)
            if any([y.endswith(ext) for ext in VALID_EXTENSIONS])]
        for candidate in candidates:
            path = os.path.join(folder, candidate)
            snap_files.append(path)

    # Enable permission to read (may not be necessary)
    workbooks = {}
    for file in snap_files:
        os.chmod(file, 0o777)
        wb = load_workbook(filename=file, read_only=True)
        workbooks[file] = wb

    worksheets = {}
    for file, wb in workbooks.items():
        ws = wb[worksheet_name]
        worksheets[file] = ws

    headers: List[str] = None
    duration_col_indices = []
    data: List[List] = []
    n_worksheets = len(worksheets)
    ws_num = 0
    for _, ws in worksheets.items():
        row_num = 0
        ws_num += 1
        if print_progress:
            print(
                'Processing file ' + str(ws_num) + ' of ' + str(n_worksheets))

        for row in ws:
            row_num += 1
            if not any(cell.value for cell in row):
                break
            row_vals = [cell.value for cell in row]
            n_cols = len(row_vals)
            if row_num == 1:
                if ws_num > 1:
                    continue
                headers = row_vals
                # identify 'duration' cols for empty test
                for i in range(n_cols):
                    if headers[i] in duration_colnames:
                        duration_col_indices.append(i)
                continue
            # check if row is 'effectively' empty
            is_empty_row = False
            if not row_vals[0]:
                for i in range(n_cols):
                    if i in duration_col_indices and \
                            row_vals[i] not in [None, 0]:
                        is_empty_row = True
                        break
                    if row_vals[i]:
                        is_empty_row = True
                        break
            if is_empty_row:
                break
            data.append(row_vals)

    df = pd.DataFrame(data, columns=headers)
    filename = (str(datetime.now()) + '.csv').replace(':', '-')
    outpath = os.path.join(output_path, filename)
    df.to_csv(outpath, index=False)
    if print_progress:
        print('Done')


if __name__ == '__main__':
    run()
